/* IR Receiver Command Interpretation Demo
 * http://frank.circleofcurrent.com/musicclock/
 * Copyright (c) 2011 Frank Zhao
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

// please see http://www.nongnu.org/avr-libc/user-manual/modules.html for info on these modules

#include <avr/io.h> // allows access to AVR built-in peripherals and registers
#include <avr/interrupt.h> // utilities to implement interrupt routines
#include <avr/pgmspace.h> // allows access to flash memory
#include <util/delay.h> // provides delays
#include <stdio.h> // allows formatted debug output
#include <avr/power.h> // used for setting clock prescaler
#include "usb_serial/usb_serial.h" // provides USB-to-serial functionality
#include <string.h>

// configuration
#define IRCTRL_PULSE_NUMOFONPULSES 34
#define IRCTRL_PULSE_STARTTHRESHOLD 8000 
#define IRCTRL_PULSE_ERROR 200 // allowed margin of error for each pulse
//#define IRCTRL_USEFLASH // do not enable, flash access is too slow
/*
 * Do not ever enable IRCTRL_USEFLASH
 * this IR command interpreter runs inside an ISR
 * through experimentation, I found that using codes stored in flash
 * memory caused the the function to run too slow for comfort
 *
 */

// pin definitions
#define IRCTRL_PORTDIR DDRD
#define IRCTRL_PORTOUTPUT PORTD
#define IRCTRL_PORTINPUT PIND
#define IRCTRL_PIN 4

// just some math macros
#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))

typedef enum { bool_false = 0, bool_true = 1, } bool_t;

typedef enum
{
    ircmd_up = 0,
    ircmd_down,
    ircmd_left,
    ircmd_right,
    ircmd_play,
    ircmd_stop,
    ircmd_mute,
    ircmd_menu,
    ircmd_none,
}
ircmd_t;

// this must be included here and only once
#include "ircodes.h"

// global variables, volatile so they can be safely used inside interrupt routines
static volatile uint16_t pulses[IRCTRL_PULSE_NUMOFONPULSES * 2];
static volatile uint8_t currentOnPulseIdx;
static volatile uint16_t lastTimestamp;
static volatile bool_t hasOverflowed;
static volatile ircmd_t lastCmd;

// this is a wrapper for usb_putchar that allows me to use printf from stdio.h
static inline int debug_putchar(char c, FILE* stream)
{
    return usb_serial_putchar(c); // refer to http://www.pjrc.com/teensy/usb_serial.html
}

// setup stream for printf from stdio.h
static FILE mystdout = FDEV_SETUP_STREAM(debug_putchar, NULL, _FDEV_SETUP_WRITE);

// reads the corresponding value stored in the code pattern memory
static inline uint16_t irctrl_getPattern(ircmd_t cmd, int idx)
{
    switch (cmd)
    {
        #ifdef IRCODE_USEFLASH
        case ircmd_up:
            return pgm_read_word(&(ircmd_pattern_up[idx]));
        case ircmd_down:
            return pgm_read_word(&(ircmd_pattern_down[idx]));
        case ircmd_left:
            return pgm_read_word(&(ircmd_pattern_left[idx]));
        case ircmd_right:
            return pgm_read_word(&(ircmd_pattern_right[idx]));
        case ircmd_play:
            return pgm_read_word(&(ircmd_pattern_play[idx]));
        case ircmd_stop:
            return pgm_read_word(&(ircmd_pattern_stop[idx]));
        case ircmd_mute:
            return pgm_read_word(&(ircmd_pattern_mute[idx]));
        case ircmd_menu:
            return pgm_read_word(&(ircmd_pattern_menu[idx]));
        #else
        case ircmd_up:
            return ircmd_pattern_up[idx];
        case ircmd_down:
            return ircmd_pattern_down[idx];
        case ircmd_left:
            return ircmd_pattern_left[idx];
        case ircmd_right:
            return ircmd_pattern_right[idx];
        case ircmd_play:
            return ircmd_pattern_play[idx];
        case ircmd_stop:
            return ircmd_pattern_stop[idx];
        case ircmd_mute:
            return ircmd_pattern_mute[idx];
        case ircmd_menu:
            return ircmd_pattern_menu[idx];
        #endif
        default:
            return 0;
    }
}

// iterate through entire history backwards
// to find a matching command by process of elimination
// returns non-valid "ircmd_code_end" if no match found
// returns valid command code if match is found
ircmd_t ircmd_check()
{
    printf("checking...\r\n");
    
    bool_t eliminated[ircmd_none];
    memset(eliminated, bool_false, ircmd_none); // clear this
    
    for (int i = 0, failedCnt = 0; i < currentOnPulseIdx && i < IRCTRL_PULSE_NUMOFONPULSES && failedCnt < ircmd_none; i++)
    { // for iterate through history of pulses
    
        for (ircmd_t k = 0; k < ircmd_none && failedCnt < ircmd_none; k++)
        { // for iterate through all commands
            
            if (eliminated[k] != bool_false)
            {
                // already eliminated, don't bother
                continue;
            }
            
            uint16_t shouldBeOnPulse = irctrl_getPattern(k, i * 2);
            uint16_t shouldBeOffPulse = irctrl_getPattern(k, (i * 2) + 1);
            
            if (pulses[i * 2] <= max((int16_t)(shouldBeOnPulse - IRCTRL_PULSE_ERROR), 0) ||
                pulses[i * 2] >= min((uint32_t)shouldBeOnPulse + IRCTRL_PULSE_ERROR, 0xFFFF) || 
                pulses[(i * 2) + 1] <= max((int16_t)(shouldBeOffPulse - IRCTRL_PULSE_ERROR), 0) ||
                pulses[(i * 2) + 1] >= min((uint32_t)shouldBeOffPulse + IRCTRL_PULSE_ERROR, 0xFFFF))
            {
                /*
                printf("eliminated %d[%d], ", k, i);
                printf("ON %d %d, ", pulses[i * 2], shouldBeOnPulse);
                printf("OFF %d %d \r\n", pulses[(i * 2) + 1], shouldBeOffPulse);
                //*/
                
                // out of range, eliminate
                eliminated[k] = bool_true;
                failedCnt++;
            }
        }
    }

    for (ircmd_t k = 0; k < ircmd_none; k++)
    { // for iterate through all commands, return the one not eliminated
        if (eliminated[k] == bool_false) return k;
    }

    return ircmd_none; // no match found, return non-valid result to indicate error
}

// output has changed, thus initiating a capture, causing this interrupt
ISR(TIMER1_CAPT_vect)
{
    // calculate using last timestamp
    uint32_t length = (((uint32_t)(ICR1)) + 0xFFFF - lastTimestamp) & 0xFFFF;
    lastTimestamp = ICR1;
    hasOverflowed = bool_false; // reset timer overflow flag
    
    if (bit_is_set(IRCTRL_PORTINPUT, IRCTRL_PIN))
    {
        // this is the end of an ON pulse
        
        if (length > IRCTRL_PULSE_STARTTHRESHOLD)
        {
            // command start
            currentOnPulseIdx = 0;
        }
        
        pulses[currentOnPulseIdx * 2] = length; // store, 0 index
        
        if (currentOnPulseIdx >= IRCTRL_PULSE_NUMOFONPULSES - 1)
        {
            ircmd_t cmd = ircmd_check();
            if (cmd != ircmd_none)
            {
                lastCmd = cmd;
                
                printf("Command: ");
            
                if (lastCmd == ircmd_up)
                {
                    printf("UP");
                }
                else if (lastCmd == ircmd_down)
                {
                    printf("DOWN");
                }
                else if (lastCmd == ircmd_left)
                {
                    printf("LEFT");
                }
                else if (lastCmd == ircmd_right)
                {
                    printf("RIGHT");
                }
                else if (lastCmd == ircmd_play)
                {
                    printf("PLAY");
                }
                else if (lastCmd == ircmd_stop)
                {
                    printf("STOP");
                }
                else if (lastCmd == ircmd_mute)
                {
                    printf("MUTE");
                }
                else if (lastCmd == ircmd_menu)
                {
                    printf("MENU");
                }
                else if (lastCmd == ircmd_none)
                {
                    printf("NONE");
                }
                
                printf("\r\n");
            }
        }
        
        // toggle trigger edge
        TCCR1B &= ~_BV(ICES1);
    }
    else
    {
        pulses[(currentOnPulseIdx * 2) + 1] = length; // store
        
        // increment pointer with wrap-around
        currentOnPulseIdx = (currentOnPulseIdx + 1) % (IRCTRL_PULSE_NUMOFONPULSES + 1);
        
        // toggle trigger edge
        TCCR1B |= _BV(ICES1);
    }
}

// timer has overflown
ISR(TIMER1_OVF_vect)
{
    if (hasOverflowed != bool_false)
    {
        // this probably means buttons has been released
        
        // consecutive overflow event without output change, timeout
        pulses[(currentOnPulseIdx * 2) + 1] = 0xFFFF;
    }
    
    hasOverflowed = 1;
}

void irctrl_init()
{
    // initialize pin as input
    IRCTRL_PORTOUTPUT &= ~_BV(IRCTRL_PIN);
    IRCTRL_PORTDIR &= ~_BV(IRCTRL_PIN);
    
    // completely reset
    hasOverflowed = bool_false;
    lastCmd = ircmd_none;
    currentOnPulseIdx = 0;
    for (int i = 0; i < IRCTRL_PULSE_NUMOFONPULSES * 2; i++)
    {
        pulses[i] = 0;
    }
    
    TCCR1B = 0b10000010; // start timer, enable input capture noise canceller, trigger on falling edge
    // prescaler is set to "div by 8"
    TIMSK1 = _BV(ICIE1) | _BV(TOIE1); // enable timer capture interrupt and timer overflow interrupt
}

int main()
{    
    clock_prescale_set(clock_div_2); // this is because we are running at 3.3V
    // see http://www.nongnu.org/avr-libc/user-manual/group__avr__power.html and http://www.pjrc.com/teensy/3volt.html
    
    stdout = &mystdout; // setup stream for printf from stdio.h
    
    usb_init(); // initialize usb obviously
    
    while (usb_configured() == 0) _delay_ms(5); // waits until USB is ready
    
    _delay_ms(5000); // 5 seconds for you to connect hyperterminal
    
    irctrl_init(); // initialize timer input capture
    sei(); // enable interrupts
    
    // here's the start of the demonstration
    printf("Begin\r\n");
    
    while (1); // infinit loop
    return 0;
}